#include "Block.h"
#include "CMacros.h"
#include "CWindow.h"
#include "LookupTables.h"
#include "CRenderer.h"
#include "Terrain.h"
#include <cassert>

using namespace std;

namespace Combustion
{
#define IN_RANGE(x, y, z) x >= 0 && x < dimensions &&\
				y >= 0 && y < dimensions &&\
				z >= 0 && z < dimensions
#define ASSERT_IN_RANGE(x, y, z) assert(IN_RANGE(x,y,x));

const float GROUND = -1.0f;
const float ISO_LEVEL = 0.0f;	//point where air/ground form a surface
const float AIR = 1.0f;

	Block::Block() : dimensions(0), pos(CVector3()),
		isInitialized(false), isEmpty(false)
	{
		data = 0;
		
		//init empty vbo/ibo pair
		vbo = CRenderer::GetInstance()->CreateVBO();
		ibo = CRenderer::GetInstance()->CreateIBO();
	}

	Block::~Block()
	{
		WipeArray();

		delete vbo;
		delete ibo;
	}

	/*
		Clear iso values for this Block
	*/
	void Block::WipeArray()
	{
		if(data != 0)
		{
			delete[] data;
		}

		data = 0;
	}

	/*
		Init sub-block within Terrain parent class (allows user to see progress
		of part of the terrain while the rest builds)
	*/
	void Block::InitBlock(Terrain* terrain, CVector3 terrainPosition,
		float block_size_xz, float block_size_y, int dimension, CVector3& position)
	{
		assert(!isInitialized);
		
		dimensions = dimension;
		pos = position;
		voxel_size_xz = block_size_xz / dimensions;
		voxel_size_y = block_size_y / dimensions;
		parent = terrain;
		terrainPos = terrainPosition;

		data = new float[dimensions * dimensions * dimensions];
	}

	/*
		Get the width/height/depth of the block
		(All Blocks are square)
	*/
	int Block::GetDimensions() 
	{ 
		return dimensions;
	}

	/*
		Get the world coordinates of this block
	*/
	CVector3 Block::GetPosition()
	{
		return pos;
	}

	/*
		Get value of cell in this Block
	*/
	float Block::GetDataAt(int col, int row, int depth)
	{
		ASSERT_IN_RANGE(row, col, depth);
		
		//x + y*WIDTH + Z*WIDTH*DEPTH
		return data[col + row * dimensions + depth * dimensions * dimensions];
	}

	/*
		Set value of cell in this block
	*/
	void Block::SetDataAt(float value, int col, int row, int depth)
	{
		ASSERT_IN_RANGE(row, col, depth);
		
		data[col + row * dimensions + depth * dimensions * dimensions] = value;
	}

	/*
		Get value of cell in this block's coordinate system
		
		If block is outside of coordinate system, ask parent terrain
		class for value of corresponding cell in the neighboring block
	*/
	float Block::GetIsoValueAtIndex(int index, int col, int row, int depth)
	{
		int X = 0, Y = 0, Z = 0;

		GetBlockLocation(X, Y, Z, index);

		col += X;	row += Y; depth += Z;

		if(IN_RANGE(col, row, depth))
			return GetDataAt(col, row, depth);
		else
			return parent->GetDataAt(int(terrainPos.X) * dimensions + col,
									 int(terrainPos.Y) * dimensions + row,
									 int(terrainPos.Z) * dimensions + depth);
	}

	/*
		Get the index into the lookup table for the cell in this block
	*/
	int Block::GetTableIndexAt(int col, int row, int depth)
	{
		int index = 0, bit = 1;	

		for(int i = 0; i <= 7; ++i)
		{
			if(GetIsoValueAtIndex(i, col, row, depth) < ISO_LEVEL) index |= bit;

			bit *= 2;
		}

		return index;
	}

	/*
		Gets the offset from a given block from a current block
		(Converts current format to the format used by the 
		 precomputed lookup table)
	*/
	void Block::GetBlockLocation(int& col, int& row, int& depth, int index)
	{
		switch(index)
		{
		case 0:		col = -1;	row =  1;	depth = -1;		break;
		case 1:		col =  1;	row =  1;	depth = -1;		break;
		case 2:		col =  1;	row =  1;	depth =  1;		break;
		case 3:		col = -1;	row =  1;	depth =  1;		break;
		case 4:		col = -1;	row = -1;	depth = -1;		break;
		case 5:		col =  1;	row = -1;	depth = -1;		break;
		case 6:		col =  1;	row = -1;	depth =  1;		break;
		case 7:		col = -1;	row = -1;	depth =  1;		break;
		default:
			CWindow::GetInstance()->DisplayMessageBox(L"Invalid Edge Lookup", L"Invalid Edge Index Provided During Edge Lookup", MB_ICONERROR);
			exit(1);
		}
	}

	/*
		Get world space coordinates for the cell in this block
	*/
	CVector3 Block::GetPositionAt(int index, int col, int row, int depth)
	{
		int X, Y, Z;
		GetBlockLocation(X, Y, Z, index);

		col   += X;
		row   += Y;
		depth += Z;

		return CVector3(voxel_size_xz * col, voxel_size_y * row, voxel_size_xz * depth);
	}

	/*
		Get the triangles needed to build the surface for this cell
		
		Includes the triangles for all neighboring cells, so skip every other
		cell when generating triangles to avoid duplicating effort
	*/
	void Block::GetTrianglesAt(int col, int row, int depth, std::map<CVertex, int, VertexMapComp>& vertMap, 
								std::vector<CVertex>& verticesList, std::vector<unsigned int>& indices, int& idxCount)
	{
		CVector3 vertices[12];

		//Get index into the edge table
		int tableIndex = GetTableIndexAt(col, row, depth);
		
		//get the value from the edge table so we can generate triangles
		//from the data
		int edges = edgeTable[ tableIndex ];

		if (edgeTable[tableIndex] == 0)
			return;
		if (edgeTable[tableIndex] & 1)
			vertices[0] = VertexInterp(ISO_LEVEL, 0, 1, col, row, depth);
		if (edgeTable[tableIndex] & 2)
			vertices[1] = VertexInterp(ISO_LEVEL, 1, 2, col, row, depth);
		if (edgeTable[tableIndex] & 4)
			vertices[2] = VertexInterp(ISO_LEVEL, 2, 3, col, row, depth);
		if (edgeTable[tableIndex] & 8)
			vertices[3] = VertexInterp(ISO_LEVEL, 3, 0, col, row, depth);
		if (edgeTable[tableIndex] & 16)
			vertices[4] = VertexInterp(ISO_LEVEL, 4, 5, col, row, depth);
		if (edgeTable[tableIndex] & 32)
			vertices[5] = VertexInterp(ISO_LEVEL, 5, 6, col, row, depth);
		if (edgeTable[tableIndex] & 64)
			vertices[6] = VertexInterp(ISO_LEVEL, 6, 7, col, row, depth);
		if (edgeTable[tableIndex] & 128)
			vertices[7] = VertexInterp(ISO_LEVEL, 7, 4, col, row, depth);
		if (edgeTable[tableIndex] & 256)
			vertices[8] = VertexInterp(ISO_LEVEL, 0, 4, col, row, depth);
		if (edgeTable[tableIndex] & 512)
			vertices[9] = VertexInterp(ISO_LEVEL, 1, 5, col, row, depth);
		if (edgeTable[tableIndex] & 1024)
			vertices[10] = VertexInterp(ISO_LEVEL, 2, 6, col, row, depth);
		if (edgeTable[tableIndex] & 2056)
			vertices[11] = VertexInterp(ISO_LEVEL, 3, 7, col, row, depth);

		CVertex triangle[3];

		/* Create Triangles */
		for(int i = 0; triTable[tableIndex][i] != -1; i += 3)
		{
			CVector3 p0 = vertices[ triTable[tableIndex][i    ] ];
			CVector3 p1 = vertices[ triTable[tableIndex][i + 1] ];
			CVector3 p2 = vertices[ triTable[tableIndex][i + 2] ];

			//each vertex normal is the face normal
			//(they'll be averaged later based on surrounding triangles
			// for smooth shading)
			CVector3 v1 = p1 - p0;
			CVector3 v2 = p2 - p0;
			CVector3 faceNormal = CVector3::CrossProduct(v1, v2);
			faceNormal.Normalize();

			triangle[0] = CVertex(p0, faceNormal, 1, 1);
			triangle[1] = CVertex(p1, faceNormal, 1, 1);
			triangle[2] = CVertex(p2, faceNormal, 1, 1);

			//remove duplicate vertices (from neighboring triangles this one shares
			//edges with) and add to the normal for smooth shading
			//(averaged later in the shader by normalizing the total)
			for(int count = 0; count < 3; ++count)
			{
				auto vertIt = vertMap.find( triangle[count] );

				if(vertIt == vertMap.end())
				{	//add unique vert to map
					indices.push_back(idxCount);
					vertMap[ triangle[count] ] = idxCount++;
					verticesList.push_back( triangle[count] );
				}
				else
				{	//vert not unique
					indices.push_back( (*vertIt).second );
					verticesList[ (*vertIt).second ].normal += faceNormal;
				}
			}
		}
	}

	/*
		Interpolate to find the correct location of the vertex in space
	*/
	CVector3 Block::VertexInterp(float isoLevel, int index1, int index2, int col, int row, int depth)
	{
		return VertexInterpolate(isoLevel, 
									GetPositionAt(index1, col, row, depth), GetPositionAt(index2, col, row, depth),
									GetIsoValueAtIndex(index1, col, row, depth), GetIsoValueAtIndex(index2, col, row, depth)
								 );
	}
	
	/*  Linearly interpolates the position where an isosurface creates
	an edge between two vertices.
	
	If P1 and P2 are the vertices of a cut edge and V1 and V2 are the scalar
	values at each vertex, the the intersection point P is given by
		
		P = P1 + (isovalue - V1) (P2 - P1) / (V2 - V1)
	*/
	CVector3 Block::VertexInterpolate(float isoValue, CVector3& p1, CVector3& p2, float pVal1, float pVal2)
	{
		float mu;
		CVector3 result;

		if( abs(isoValue - pVal1) < 0.00001)
			return p1;
		else if( abs(isoValue - pVal2) < 0.00001)
			return p2;
		else if( abs(pVal1 - pVal2) < 0.00001)
			return p1;

		mu = (isoValue - pVal1) / (pVal2 - pVal1);
		result.X = p1.X + mu * (p2.X - p1.X);
		result.Y = p1.Y + mu * (p2.Y - p1.Y);
		result.Z = p1.Z + mu * (p2.Z - p1.Z);

		return result;
	}

	/*
		Generate triangles for the iso data in this block.
		Generate vbo and ibo for the triangles.
	*/
	void Block::InitBuffers()
	{
		if(isEmpty) return;	//already processed voxel grid and found no triangles

		map<CVertex, int, VertexMapComp> vertMap;	//maps vertex to index
		vector<CVertex> vertices;
		vector<unsigned int> indices;
		int idxCount = 0;

		//special case for edges (ground to empty space creates large surfaces at the edges)
		//starting at 2 so the +=2 matches up
		int startI = (int(terrainPos.X) == 0) ? 2 : 0;
		int startJ = (int(terrainPos.Y) == 0) ? 2 : 0;
		int startK = (int(terrainPos.Z) == 0) ? 2 : 0;

		for(int i = startI; i < dimensions; i += 2) {
		    for(int j = startJ; j < dimensions; j += 2) {
		        for(int k = startK; k < dimensions; k += 2) {
					GetTrianglesAt(i, j, k, vertMap, vertices, indices, idxCount);
				}
			}
		}

		if(vertices.size() == 0 || indices.size() == 0) 
		{
			isEmpty = true;
		}
		else
		{
			vbo->GenerateVBO(vertices);
			ibo->CreateIBO(indices);
		}
	}
}